@@ -1,5 +1,5 @@
 /* dfasearch.c - searching subroutines using dfa and regex for grep.
-   Copyright 1992, 1998, 2000, 2007, 2009-2011 Free Software Foundation, Inc.
+   Copyright 1992, 1998, 2000, 2007, 2009-2012 Free Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -19,6 +19,7 @@
 /* Written August 1992 by Mike Haertel. */
 
 #include <config.h>
+#include "intprops.h"
 #include "search.h"
 #include "dfa.h"
 
@@ -71,7 +72,7 @@ dfawarn (char const *mesg)
 /* Number of compiled fixed strings known to exactly match the regexp.
    If kwsexec returns < kwset_exact_matches, then we don't need to
    call the regexp matcher at all. */
-static int kwset_exact_matches;
+static size_t kwset_exact_matches;
 
 static char const *
 kwsincr_case (const char *must)
@@ -211,7 +212,9 @@ EGexecute (char const *buf, size_t size, size_t *match_size,
 {
   char const *buflim, *beg, *end, *match, *best_match, *mb_start;
   char eol = eolbyte;
-  int backref, start, len, best_len;
+  int backref;
+  regoff_t start;
+  ptrdiff_t len, best_len;
   struct kwsmatch kwsm;
   size_t i, ret_val;
   if (MB_CUR_MAX > 1)
@@ -294,6 +297,11 @@ EGexecute (char const *buf, size_t size, size_t *match_size,
           end = buflim;
         }
 
+      /* If the "line" is longer than the maximum regexp offset,
+         die as if we've run out of memory.  */
+      if (TYPE_MAXIMUM (regoff_t) < end - buf - 1)
+        xalloc_die ();
+
       /* If we've made it to this point, this means DFA has seen
          a probable match, and we need to run it through Regex. */
       best_match = end;
@@ -301,10 +309,13 @@ EGexecute (char const *buf, size_t size, size_t *match_size,
       for (i = 0; i < pcount; i++)
         {
           patterns[i].regexbuf.not_eol = 0;
-          if (0 <= (start = re_search (&(patterns[i].regexbuf),
-                                       buf, end - buf - 1,
-                                       beg - buf, end - beg - 1,
-                                       &(patterns[i].regs))))
+          start = re_search (&(patterns[i].regexbuf),
+                             buf, end - buf - 1,
+                             beg - buf, end - beg - 1,
+                             &(patterns[i].regs));
+          if (start < -1)
+            xalloc_die ();
+          else if (0 <= start)
             {
               len = patterns[i].regs.end[0] - start;
               match = buf + start;
@@ -321,7 +332,7 @@ EGexecute (char const *buf, size_t size, size_t *match_size,
                 }
               /* If -w, check if the match aligns with word boundaries.
                  We do this iteratively because:
-                 (a) the line may contain more than one occurence of the
+                 (a) the line may contain more than one occurrence of the
                  pattern, and
                  (b) Several alternatives in the pattern might be valid at a
                  given point, and we may need to consider a shorter one to
@@ -341,6 +352,8 @@ EGexecute (char const *buf, size_t size, size_t *match_size,
                         len = re_match (&(patterns[i].regexbuf),
                                         buf, match + len - beg, match - buf,
                                         &(patterns[i].regs));
+                        if (len < -1)
+                          xalloc_die ();
                       }
                     if (len <= 0)
                       {
@@ -354,7 +367,11 @@ EGexecute (char const *buf, size_t size, size_t *match_size,
                                            match - buf, end - match - 1,
                                            &(patterns[i].regs));
                         if (start < 0)
-                          break;
+                          {
+                            if (start < -1)
+                              xalloc_die ();
+                            break;
+                          }
                         len = patterns[i].regs.end[0] - start;
                         match = buf + start;
                       }
