@@ -1,5 +1,5 @@
 /* kwset.h - header declaring the keyword set library.
-   Copyright (C) 1989, 1998, 2005, 2007, 2009-2010 Free Software Foundation,
+   Copyright (C) 1989, 1998, 2005, 2007, 2009-2012 Free Software Foundation,
    Inc.
 
    This program is free software; you can redistribute it and/or modify
@@ -23,11 +23,13 @@
 
 struct kwsmatch
 {
-  int index;			/* Index number of matching keyword. */
+  size_t index;			/* Index number of matching keyword. */
   size_t offset[1];		/* Offset of each submatch. */
   size_t size[1];		/* Length of each submatch. */
 };
 
+#include "arg-nonnull.h"
+
 struct kwset;
 typedef struct kwset *kwset_t;
 
@@ -52,8 +54,8 @@ extern const char *kwsprep (kwset_t);
    the matching substring in the integer it points to.  Similarly,
    if foundindex is non-NULL, store the index of the particular
    keyword found therein. */
-extern size_t kwsexec (kwset_t, char const *, size_t, struct kwsmatch *);
+extern size_t kwsexec (kwset_t, char const *, size_t, struct kwsmatch *)
+  _GL_ARG_NONNULL ((4));
 
 /* Deallocate the given keyword set and all its associated storage. */
 extern void kwsfree (kwset_t);
-
